# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# Boiler Control 

"""
Boiler Control Unit (BCU)

The BCU  has a button to switch between solar (S) or grid (G) powered heater power supply

S)   Full power supply, not regarding wether there is enough solar power or not
G)   No grid power consumption. The heater power depends on the available solar power
       Consumption adjustment is done by controlling the three 230V heater coils
       - control phase L1 between 0 and 100%
       - switch on / off the L2 and/or  L3 heater 230V phases

The Central Control Unit (CCU,  RasPi) is connected to the solar edge power meter, reads the current grid/solar power values
and provides the commands for the BCU.

To do so  the BCU receives  a command from  the Central Control Unit( CCU )
transmitted with LoRa and responds the command.

Currently, the CCU sends the available solar power and asks for the boiler temperature ( DS18B20) and
the ambient temperature and humidity (DHT22).
The DS18B20 is installed within the boiler but has no direct contact to the hot water.
Therefore, a temperature deviation is likely and may be adjusted  by a fix number of degrease in the CCU

"""
from onewire import OneWire
from ds18x20 import DS18X20
import dht

from time import sleep
from machine import Pin, PWM, Timer, ADC
from neopixel import NeoPixel
from random import randint
import config as cfg
from tools import LoRaTools


class BoilerControl:
    def __init__(self, spi):
        self.npix = NeoPixel(Pin(cfg.Pin["NeoPixel"], Pin.OUT), 12)  # create NeoPixel driver on GPIOx for 12 pixels
        self.pwm_L1 = PWM(Pin(cfg.Pin["PhaseL1"]))  # , freq=1000)
        self.pin_L2 = Pin(cfg.Pin["PhaseL2"], Pin.OUT)
        self.pin_L3 = Pin(cfg.Pin["PhaseL3"], Pin.OUT)
        self.btn_auto = Pin(cfg.Pin["ButtonAutomatic"], Pin.IN, Pin.PULL_DOWN)
        self.btn2 = Pin(cfg.Pin["Button2"], Pin.IN, Pin.PULL_DOWN)
        self.pin_on_off = Pin(cfg.Pin["BoilerOnOff"], Pin.IN, Pin.PULL_DOWN)
        self.ow = OneWire(Pin(cfg.Pin["DS18B20"]))  # create a OneWire bus
        self.thd = dht.DHT22(Pin(cfg.Pin["DHT22"]))
        self.ldr_adc = ADC(Pin(cfg.Pin["LDR"]))

        self.pix_red = []
        self.pix_green = []
        self.pix_blue = []
        self.pix_white = []
        self.pix_yellow = []
        self.pix_cyan = []
        self.pix_magenta = []
        self.pix_black = []

        self.rx_msg = "-"
        self.no_command_count = 1
        self.L1_pwm_power = 0
        self.opmode_solar = True  # False = Grid Mode
        self.loratools = LoRaTools(spi)
        self.tim = Timer()
        self.power_up()
        self.loratools.init_recv_mode(self.receive)
        self.btn_auto.irq(trigger=Pin.IRQ_RISING, handler=self.btn_pressed)
        # self.check_hardware(10)
        self.tim.init(period=1000 * cfg.BASIC_TICK, callback=self.check_connection)  # msec

    def check_hardware(self, repetition):
        """ For test purposes and new sensor installation"""
        for i in range(repetition):
            print(76, "HW-Check: ", i, "/", repetition)
            roms = self.ow.scan()
            for rom in roms:
                print(rom)  # rom ids from all 1wire devices as bytearray

            print(self.read_boiler_temp())     # temperature
            print(self.read_ambient_temp())  # temperature, humidity

            read = self.ldr_adc.read_u16()
            print(85, "ADC: ", read)
            print(86, "BoilerOnOff: ", self.pin_on_off.value())
            sleep(2)

    def send(self, tx_data):
        """
        Completes and sends a message , (re)enables the reception mode
        tx_data structure:  BCU_ID+CCU_ID+Command+Heater_On/Off+Grid/Solar+Values
        """
        # Add Transmitter and Receiver ID
        tx_data = cfg.LoRaNode["Boiler"] + cfg.LoRaNode["F40CentralUnit"] + tx_data
        print(96, "TX ", tx_data)
        self.loratools.send(tx_data)
        self.loratools.set_recv_mode()

    def receive(self, pin_irq):
        """ This function is triggered ( DIO0 IRQ) by a new LoRa  message """
        self.process_rxdata(self.loratools.get_recv_data())

    def process_rxdata(self, rx_data):
        """
         Processes the received message 
         and sends an acknowledgement back
          """
        print(109, "RX ", rx_data)
        if not self.rx_check(rx_data):
            return
        self.rx_msg = rx_data
        sleep(1.5)  # Don't response too quick for my Raspi 2B+

        if self.rx_msg[2] == cfg.Cmd["set_power"]:
            power = int(self.rx_msg[3:])
            self.activate_heater(power)
            tx_msg = cfg.Cmd["set_power"] + self.update_opmode() + self.rx_msg[3:]   # skip sender/receiver tag in the received message
        elif self.rx_msg[2] == cfg.Cmd["get_sonsordata"]:
            temp, humi = self.read_ambient_temp()
            boiler_temperature = self.read_boiler_temp()
            stext = "{:0>4.1f},{:0>4.1f},{:0>4.1f}".format(boiler_temperature, temp, humi)
            tx_msg = cfg.Cmd["get_sonsordata"] + self.update_opmode() + stext
        else:
            print(125, "Command not implemented", self.rx_msg[2])
            return
        self.send(tx_msg)  # send response

    def rx_check(self, rx_data):
        self.no_command_count = 1
        self.switch_neopix_led(cfg.NeopixSettings["LedHostConnection"], color=self.pix_green, pixon=True)
        if rx_data == cfg.ERR_ERROR:  # error, root is the lora payload function
            return False
        elif len(rx_data) < 3:
            return False
        elif rx_data[2] == cfg.ALIVE_SIGNAL_PREFIX:  # except alive signal from every client
            print(112, "RX Alive signal", rx_data)
            return False
        elif rx_data[0] != cfg.LoRaNode["F40CentralUnit"]:  # not my boss
            return False
        elif rx_data[1] != cfg.LoRaNode["Boiler"]:  # not for me
            return False
        return True

    def activate_heater(self, power):
        """
        Find heating coil settings
        """
        if not self.opmode_solar:
            power = cfg.HEATER_POWER_MAX
        self.L1_pwm_power = 0
        if power >= cfg.SWITCH_POINT_HIGH + cfg.L1_POWER_MIN:  # 4000W + 200W
            self.L1_pwm_power = power - cfg.SWITCH_POINT_HIGH
            self.heater_onoff(2, on=True)
            self.heater_onoff(3, on=True)
        elif power >= cfg.SWITCH_POINT_LOW + cfg.L1_POWER_MIN:
            hid = randint(2, 3)  # avoids using always the same phase
            self.L1_pwm_power = power - cfg.SWITCH_POINT_LOW
            if hid == 2:  # use phase1 or phase2 coincidentally
                self.heater_onoff(2, on=True)
                self.heater_onoff(3, on=False)
            else:
                self.heater_onoff(2, on=False)
                self.heater_onoff(3, on=True)
        elif power < cfg.SWITCH_POINT_LOW:
            self.L1_pwm_power = power
            self.heater_onoff(2, on=False)
            self.heater_onoff(3, on=False)
        self.pwm_L1.duty_u16(32 * self.L1_pwm_power)  # 0 .. 2000 --> 0 .. 65532
        self.set_phase_neopixel()

    def heater_onoff(self, hid, on=False):
        if hid == 2:
            self.pin_L2.value(on)
        else:  # hid = 3
            self.pin_L3.value(on)

    def get_phase_status(self):
        return self.L1_pwm_power, self.pin_L2.value(), self.pin_L3.value()

    def stop_pwm(self):
        self.pwm_L1.deinit()

    def set_phase_neopixel(self):
        """
        Set Neopixel LEDs
        """
        # switch off phase neopixel
        for i in cfg.NeopixSettings["LedPhase1"]:
            self.switch_neopix_led(i, pixon=False)
        self.switch_neopix_led(cfg.NeopixSettings["LedPhase2"], pixon=False)
        self.switch_neopix_led(cfg.NeopixSettings["LedPhase3"], pixon=False)

        l1_power, l2, l3 = self.get_phase_status()
        if l1_power > 0:
            self.switch_neopix_led(cfg.NeopixSettings["LedPhase1"][0], color=self.pix_cyan, pixon=True)
        if l1_power > 500:
            self.switch_neopix_led(cfg.NeopixSettings["LedPhase1"][1], color=self.pix_cyan, pixon=True)
        if l1_power > 1000:
            self.switch_neopix_led(cfg.NeopixSettings["LedPhase1"][2], color=self.pix_cyan, pixon=True)
        if l1_power > 1500:
            self.switch_neopix_led(cfg.NeopixSettings["LedPhase1"][3], color=self.pix_cyan, pixon=True)
        self.switch_neopix_led(cfg.NeopixSettings["LedPhase2"], color=self.pix_cyan, pixon=l2)
        self.switch_neopix_led(cfg.NeopixSettings["LedPhase3"], color=self.pix_cyan, pixon=l3)

    def switch_neopix_led(self, pix_number, color=cfg.NeopixSettings["Blue"], pixon=None):
        if pixon:
            self.npix[pix_number] = color
        else:
            self.npix[pix_number] = self.pix_black
        self.npix.write()

    def power_up(self):
        self.adjust_neopixel_brightness()
        for i in range(0, 12):
            self.switch_neopix_led(i, pixon=True)
            sleep(0.3)
            self.switch_neopix_led(i, pixon=False)
        for i in cfg.NeopixSettings["LedNotUsed"]:
            self.switch_neopix_led(i, color=self.pix_black, pixon=True)
        self.opmode_solar = True  # False = Grid Mode
        self.switch_neopix_led(cfg.NeopixSettings["LedHostConnection"], color=self.pix_red, pixon=True)
        self.update_opmode()
        self.activate_heater(0)

    def btn_pressed(self, pin):
        """
        Set solar or grid mode.
        """
        self.btn_auto.irq(trigger=0)        # Disable IRQ , avoids debouncing
        if pin == self.btn_auto:
            while self.btn_auto.value():
                # Toogle between operation modes with a delay
                self.switch_neopix_led(cfg.NeopixSettings["LedGridSolar"], color=self.pix_yellow, pixon=True)
                self.opmode_solar = True
                self.activate_heater(0)
                self.set_phase_neopixel()
                sleep(2)
                if not self.btn_auto.value():
                    break
                self.switch_neopix_led(cfg.NeopixSettings["LedGridSolar"], color=self.pix_magenta, pixon=True)
                self.opmode_solar = False
                self.activate_heater(cfg.HEATER_POWER_MAX)
                sleep(2)
        elif pin == self.btn2:
            pass
        # Re-enable IRQ
        self.btn_auto.irq(trigger=Pin.IRQ_RISING, handler=self.btn_pressed)

    def check_connection(self, timerobj):
        """
        This function is called by a periodic timer, updates the neopixels and checks the host connection.
        Indicator is the self.no_command_count variable.
        """
        # do some miscellaneous actions
        self.adjust_neopixel_brightness()  # set day/night neopixel brightness
        self.set_phase_neopixel()
        self.update_opmode()   # updates opmode in offline mode
      
        if self.no_command_count == 0:      # Data connection error
            return
        checktime = cfg.BASIC_TICK * self.no_command_count
        self.no_command_count += 1  # increase check intervall
        if checktime >= cfg.HOST_CONNECTION_TIMEOUT:
            self.switch_neopix_led(cfg.NeopixSettings["LedHostConnection"], color=self.pix_red, pixon=True)
            self.no_command_count = 0  # Data connection error indicator
            self.send_alive()
            if self.opmode_solar:
                self.activate_heater(0)
                print(272, "No command received from the CCU since  ", checktime, " sec")

    def adjust_neopixel_brightness(self):
        """Get an ambient light value and adjust the Neopixel brightness"""
        brightness = 1
        ldr_value = self.ldr_adc.read_u16()  # read a raw analog value in the range 0-65535
        # print(286, "LDR-ADC: ", ldr_value)
        if ldr_value < cfg.LDRSwitchPoint["Night"]:
            brightness = cfg.NeopixSettings["NightBrightness"]
        elif ldr_value > cfg.LDRSwitchPoint["Day"]:
            brightness = cfg.NeopixSettings["DayBrightness"]
        self.pix_red = [x * brightness for x in cfg.NeopixSettings["Red"]]
        self.pix_green = [x * brightness for x in cfg.NeopixSettings["Green"]]
        self.pix_blue = [x * brightness for x in cfg.NeopixSettings["Blue"]]
        self.pix_white = [x * brightness for x in cfg.NeopixSettings["White"]]
        self.pix_yellow = [x * brightness for x in cfg.NeopixSettings["Yellow"]]
        self.pix_cyan = [x * brightness for x in cfg.NeopixSettings["Cyan"]]
        self.pix_magenta = [x * brightness for x in cfg.NeopixSettings["Magenta"]]
        self.pix_black = [x * brightness for x in cfg.NeopixSettings["Black"]]
        self.set_pix_hostconnection(brightness)

    def set_pix_hostconnection(self, brightness):
        """ Reads the LedHostConnection color tuple and writes it back with updated brightness."""
        color = self.npix.__getitem__(cfg.NeopixSettings["LedHostConnection"])
        color = [x * brightness for x in color]
        self.switch_neopix_led(cfg.NeopixSettings["LedHostConnection"], color=color, pixon=True)

    def send_alive(self):
        txdata = cfg.ALIVE_SIGNAL_PREFIX + " Alive"
        self.send(txdata)

    def read_boiler_temp(self):
        """  OneWire sensor DS18B20 """
        ds_sensor = DS18X20(self.ow)
        ds_sensor.convert_temp()
        sleep(0.750)

        try:
            temperature = ds_sensor.read_temp(b'(\xab\xe5)\t\x00\x00\xdf')
        except Exception as e:
            print("314", "Can't read boiler sensor DS18B20.  Sensor-ID ok? ", e)
            return cfg.ERR_ERROR
        return temperature

    def read_ambient_temp(self):
        """ DHT22 temperature and humidity sensor """
        self.thd.measure()
        sleep(0.5)
        return self.thd.temperature(), self.thd.humidity()

    def update_opmode(self):
        # Heater on / off
        if self.pin_on_off.value():   # Boiler current on Phase L1 detected
            opmode = cfg.HeaterOpMode["HeaterOn"]
            self.switch_neopix_led(cfg.NeopixSettings["LedBoilerOnOff"], color=self.pix_white, pixon=True)
        else:
            opmode = cfg.HeaterOpMode["HeaterOff"]
            self.switch_neopix_led(cfg.NeopixSettings["LedBoilerOnOff"], color=self.pix_black, pixon=False)
        # Solar / Grid powered
        if self.opmode_solar:
            opmode = opmode + cfg.HeaterOpMode["SolarPowered"]
            self.switch_neopix_led(cfg.NeopixSettings["LedGridSolar"], color=self.pix_yellow, pixon=True)
        else:
            opmode = opmode + cfg.HeaterOpMode["GridPowered"]
            self.switch_neopix_led(cfg.NeopixSettings["LedGridSolar"], color=self.pix_magenta, pixon=True)
        return opmode


if __name__ == "__main__":
    exec(open("boiler_run.py").read())
