# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# Pico tools


import config as cfg
from machine import Pin, SPI
from time import sleep
from lora_upy import LoRa


class BoilerTools:
    def __init__(self):
        pass

    def init_spi(self):
        spi = SPI(cfg.LoRaSpi["Bus"],
                  baudrate=cfg.LoRaSpi["Baudrate"],
                  polarity=cfg.LoRaSpi["Polarity"],
                  phase=cfg.LoRaSpi["Phase"],
                  sck=Pin(cfg.LoRaSpi["PinSCK"], Pin.OUT),
                  mosi=Pin(cfg.LoRaSpi["PinMOSI"], Pin.OUT),
                  miso=Pin(cfg.LoRaSpi["PinMISO"], Pin.IN))
        spi.init()
        return spi


class LoRaTools:

    def __init__(self, spi):
        self.pin_nss = Pin(cfg.LoraPins["NSS"], Pin.OUT)
        self.pin_rst = Pin(cfg.LoraPins["Reset"], Pin.OUT, Pin.PULL_UP)
        self.pin_rxirq = Pin(cfg.LoraPins["DIO0"], Pin.IN)
        self.spi = spi
        self.lora_reset()
        self.lora = LoRa(self.spi, self.get_lorapins())
        self.irq_dio0 = None
        self.show_modem_config()

    def get_lorapins(self):
        return self.pin_rst, self.pin_nss, self.pin_rxirq

    def init_spi(self):
        spi = SPI(cfg.LoRaSpi["Bus"],
            baudrate=cfg.LoRaSpi["Baudrate"],
            polarity=cfg.LoRaSpi["Polarity"],
            phase=cfg.LoRaSpi["Phase"],
            sck=Pin(cfg.LoRaSpi["PinSCK"], Pin.OUT),
            mosi=Pin(cfg.LoRaSpi["PinMOSI"], Pin.OUT),
            miso=Pin(cfg.LoRaSpi["PinMISO"], Pin.IN))
        spi.init()
        return spi

    def lora_reset(self):
        """ Reset LoRa device, do a reset pulse """
        self.pin_rst.value(1)
        sleep(.01)
        self.pin_rst.value(0)
        sleep(.01)
        self.pin_rst.value(1)
        sleep(.01)

    def cancel_recv(self):
        """ Cancel  pin  interrupt """
        self.pin_rxirq.irq(handler=None)

    def init_recv_mode(self, callback):
        """
        Start the Raspi /ESP/Pico  Rx  interrupt and wait for incoming messages .
        callback is a handle, pointing to the function that will be processed when an interrupt occurs.
        """
        self.irq_dio0 = self.pin_rxirq.irq(handler=callback, trigger=Pin.IRQ_RISING)
        self.lora.set_rx_mode()

    def send(self, tx_data):
        self.lora.send(tx_data)

    def set_recv_mode(self):
        self.lora.set_rx_mode()

    def get_recv_data(self):
        return self.lora.get_rx_data()

    def get_irq(self):
        return self.irq_dio0

    def show_register(self, line_no=00, register=0x01, text="--"):
        """ line_no:  line number of the python script"""
        xx = self.lora.read_register(register)
        print(line_no, text, "0x{:02x}\t{}\t0x{:02x}\t{:08b}".format(register, xx, xx, xx))  # decimal, hex, binary

    def read_modem_status(self):
        ms = self.lora.read_modem_status()
        return dict(
            rx_coding_rate=ms >> 5 & 0x03,
            modem_clear=ms >> 4 & 0x01,
            header_info_valid=ms >> 3 & 0x01,
            rx_ongoing=ms >> 2 & 0x01,
            signal_sync=ms >> 1 & 0x01,
            signal_detected=ms >> 0 & 0x01
        )

    def read_irq_flags(self):
        iqf = self.lora.read_irq_flags()
        return dict(
            rx_timeout=iqf >> 7 & 0x01,
            rx_done=iqf >> 6 & 0x01,
            crc_error=iqf >> 5 & 0x01,
            iqvalid_header=iqf >> 4 & 0x01,
            tx_done=iqf >> 3 & 0x01,
            cad_done=iqf >> 2 & 0x01,
            fhss_change_ch=iqf >> 1 & 0x01,
            cad_detected=iqf >> 0 & 0x01,
        )

    def read_operation_mode(self):
        om = self.lora.read_operation_mode()
        return dict(
            LongRangeMode=om >> 7 & 0x01,      # 7
            AccessSharedReg=om >> 6 & 0x01,    # 6
            reserved=om >> 4 & 0x03,   # 5-4
            LowFrequencyModeOn=om >> 3 & 0x01,     # 3
            mode=om >> 0 & 0x07,       # 2-0
        )

    def debug(self, line, irq_dio0=None):
        print(line, self.read_operation_mode())
        print(line, self.read_irq_flags())
        print(line, self.read_modem_status())

    def show_modem_config(self):
        self.show_register(line_no=00, register=0x1d, text="REG_MODEM_CONFIG_1")
        self.show_register(line_no=00, register=0x1e, text="REG_MODEM_CONFIG_2")
        self.show_register(line_no=00, register=0x26, text="REG_MODEM_CONFIG_3")
