# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220402
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# Surplus energy management

from time import sleep
import config as cfg
from tools import Misc as MiscTools
import threading
from os import path


class HeaterControl:
    def __init__(self, logger, oledtools, loratools, solar_registers):

        self.filename = path.split(__file__)[1]
        self.logger = logger
        if not self.logger:
            print(self.filename, "ERROR No logger defined. Exit programm")
            exit(0)
        self.logger.info("File: {} ; {} ; {} ; {}".format(self.filename, "Heater control started", "--", "--"))

        self.oledtools = oledtools
        self.oledtools.oled_init()
        self.oledtools.oled_startup_screen()
        self.stack = ["----",  "----",  "----", "----"]    # Oled scroll stack
        self.lora_id = cfg.LoRaNode["F40CentralUnit"] + cfg.LoRaNode["Boiler"]
        self.loratools = loratools
        self.command_last = "-"
        self.ac_power_last = 0
        self.heater_on_last = False
        self.rx_msg = "-"
        self.solarregs = solar_registers
        self.is_alive_count = 0
        self.is_alive_logtrigger = cfg.NO_RESPONSE_MAX
        self.msg_count = 0
        self.heaterpower_solar = True
        self.heater_on = True
        sleep(2)  # don't hurry, wait for Modbus and LoRa
        self.data_update_count = 0
        self.sorted_updatedict, self.data_update_count_max = MiscTools().sort_dict(cfg.DataUpdate)
        self.loratools.init_recv_mode(self.receive)
        self.trigger_event()

    def trigger_event(self, threadfiretime=cfg.BASIC_TICK):
        """
        Calls every threadfiretime (sec) for new power and sensor data
        The complete update process is "slow", ensure that the received messages of different
        updates do not overhaul. (config file , DataUpdate dictionary)
         """
        self.data_update_count += 1
        event_key = "-"
        for item in self.sorted_updatedict.items():
            x = self.data_update_count % item[1]
            if x == 0:
                event_key = item[0]
                self.update_data(event_key)
                break
        if self.data_update_count == self.data_update_count_max:
            self.data_update_count = 0

        threading.Timer(threadfiretime, self.trigger_event).start()
        return

    def update_data(self, event_key):
        """
        Calls every threadfiretime (sec) for new power and sensor data
        The complete update process is "slow", ensure that the received messages of different
        updates do not overhaul. (config file , DataUpdate dictionary)
         """

        if event_key == "SolarData":
            if self.solarregs.read_holding_regs() == cfg.ERR_ERROR:
                self.data_update_count -= 1  # Repeat reading next time
                self.logger.warning("File: {} ; ModbusTCP {} ; Reading Solar Registers failed {} ; {}".format(self.filename, "--", "--", "--",))
        elif event_key == "Power":
            self.power_update(event_key)  # solar plant data
        elif event_key == "SensorData":
            self.command_last = cfg.Cmd["get_sonsordata"]
            self.wait_for_response(self.command_last, event_key)
        elif event_key == "ArchiveData":
            self.solarregs.archive()
        return

    def send(self, tx_data):
        """
        Add transmitter and receiver ID and send the message
        """
        self.msg_count += 1
        tx_data = self.lora_id + tx_data
        text = "{:04d} {} {} {}".format(self.msg_count, tx_data[:2], tx_data[2], tx_data[3:])
        # print("81", "TX", text)
        self.oledtools.oled_scroll(self.stack, text)

        self.logger.info("File: {} ; MsgCount: {} ; TX: {} ; Oled-Stack {}".format(
            self.filename,
            self.msg_count,
            tx_data,
            self.stack,
            )
        )
        self.loratools.send(tx_data)

    def receive(self, pin, level, tick):
        """ This function is triggered ( DIO0 IRQ) by a new LoRa  message """
        self.process_rxdata(self.loratools.get_recv_data())

    def send_receive(self, msg):
        """  Sends a LoRa message  and (re)enables the reception mode  """
        self.rx_msg = "-"
        if self.is_alive_count >= cfg.NO_RESPONSE_MAX:
            self.set_alive_log()
        self.send(msg)
        self.is_alive_count += 1
        self.loratools.set_recv_mode()

    def process_rxdata(self, rx_data):
        """
        Checks the received data validity and processes the data,
        depending on the command id
        """
        check_msg = self.rx_check(rx_data)
        if check_msg is not None:
            self.logger.warning("File: {} ; RX {} ; {} ; {}".format(
                        self.filename,
                        self.rx_msg,
                        "--",
                        check_msg)
                        )
            self.rx_msg = str(cfg.ERR_ERROR)
            self.loratools.set_recv_mode()
            return

        packet_rssi_strength = str(self.loratools.get_packet_rssi_strength())
        text = "RX {} {}".format(rx_data, packet_rssi_strength)
        self.oledtools.oled_scroll(self.stack, text)

        self.logger.info("File: {} ; {} ; RX: {} ; Signal Strength(dbm): {}".format(        # info
            self.filename,
            "--",
            rx_data,
            packet_rssi_strength,
            )
        )

        self.command_last = "-"
        self.rx_msg = rx_data
        self.logger.debug("File: {} ; RX {} ;  {} ; {}".format(
            self.filename,
            self.rx_msg,    # self.rx_msg[3:].split(","),
            "--",
            "--",
            )
        )
        self.loratools.set_recv_mode()

    def rx_check(self, rx_data):
        """
        Simple check of the received data. They are not qeued or managed.
        Time delays ( config.py --> DataUpdate ) between two command must ensure that the
        received messages do not overtake each other
        """
        if type(rx_data) == str:
            self.rx_msg = ""
            for char in rx_data:
                if not char.isprintable():
                    char = "*"
                self.rx_msg += char
            if rx_data != self.rx_msg:
                self.logger.warning("File: {} ; RX  {} ; Corrupted characters replaced by: {} ; {}".format(
                    self.filename,
                    self.rx_msg,
                    "*",
                    "--")
                )
                return "RX Corrupted characters"
            return

        self.rx_msg = rx_data
        if rx_data == cfg.ERR_ERROR:  # error, root is the lora payload function
            return "Lora payload CRC error"
        if len(rx_data) < 3:
            return "RX data too short"
        elif rx_data[1] != cfg.LoRaNode["F40CentralUnit"]:  # not for me
            return "RX message not for me"
        elif rx_data[2] == cfg.ALIVE_SIGNAL_PREFIX:  # except alive signal from every client
            self.set_alive(rx_data)
            return
        elif rx_data[0] != cfg.LoRaNode["Boiler"]:  # not my client
            return "RX transmitter is not my client"
        elif rx_data[2] != self.command_last[0]:
            return "RX unexpected command"
        elif len(rx_data) < 6:  # Client_ID + CCU_ID + Command + ON/Off + Solar/Grid + Value
            return "rx data not complete"

        if rx_data[2] == cfg.Cmd["set_power"]:
            self.heater_on_last = False
            if rx_data[3] == cfg.HeaterOpMode["HeaterOn"]:
                self.heater_on_last = True
            if rx_data[5:] != str(self.ac_power_last):
                self.logger.error("File: {} ; Heater On? {} ; Power Deviation: TX {} ; RX {}".format(
                    self.filename,
                    self.heater_on_last,
                    self.ac_power_last,
                    rx_data[5:])
                    )
                return "RX unexpected power value"

        # A valid response resets the alive parameter
        self.is_alive_count = 0
        print(211, "alive_count: ", self.is_alive_count)
        self.is_alive_logtrigger = cfg.NO_RESPONSE_MAX
        return

    def set_alive(self, rx_data):
        """
        Logs alive signals.
        Mainly for test purposes with the ESP32-TTGO Lora sniffer.
        """
        self.logger.debug("File: {} ; {} ; Alive received {} ; {}".format(
            self.filename,
            "--",
            rx_data,
            "--")
            )

    def set_alive_log(self):
        """
        Increase the logger message intervall if the client responses fails to appear for long times
        The time between 2 logs depends on the cfg.DataUpdate["BaseInterval"] entry
        """
        if self.is_alive_count % self.is_alive_logtrigger == 0:
            self.logger.critical("File: {} ; {} ; {} ; {}".format(
                self.filename,
                "Boiler alive? No response since: ",
                self.is_alive_count,
                "  requests")
                )
            self.is_alive_logtrigger *= 2  # increase log intervall

    def power_update(self, txt):
        """
        Reads the  exported power from the solar plant meter,
        calculates the available heater power,  sends the power to the
        heater BCU and waits for an acknowledgement
        """
        ac_power = self.solarregs.get_heater_power()
        self.power_status(ac_power)
        if ac_power < cfg.AC_POWER_MIN:  # otherwise  fast on/offs are  likely
            ac_power = 0
        if ac_power > cfg.AC_POWER_MAX:  # not necessary, the power is limited by the heater itself
            ac_power = cfg.AC_POWER_MAX

        # ac_power = 1111
        msg = cfg.Cmd["set_power"] + str(ac_power)
        self.command_last = cfg.Cmd["set_power"]
        self.ac_power_last = ac_power
        self.wait_for_response(msg, txt)

    def power_status(self, ac_power):
        """
        If on, it's assumed that the heater consumes the "self.ac_power_last" power.
        The new  power request reduces (-ac_power, import from grid) or increases (+ac_power, export to grid)
        the total available power for the heater
        """
        if self.heater_on_last:
            ac_power += self.ac_power_last  # total available power
            return
        return

    def wait_for_response(self, msg, datatext):
        """
        Send a message and wait for the response.
        Loop/time figured out by trial & error
        """
        wait_loop = 0
        self.send_receive(msg)
        self.reg_status(284)
        for wait_loop in range(10001):
            if self.rx_msg != "-":
                self.reg_status(280)
                return
            sleep(0.001)
        self.logger.warning("File: {} ; RX Timeout {}msec ; No {} response; {}".format(
            self.filename,
            wait_loop,
            datatext,
            msg)
            )
        self.reg_status(298)
        self.loratools.set_recv_mode()
        self.reg_status(300)

    def reg_status(self, line):
        dicta = self.loratools.read_operation_mode()
        dictb = self.loratools.read_modem_status()
        dictc = self.loratools.read_irq_flags()
        self.logger.debug("File: {} ; Line: {} : RegModemStatus(0x18) {} ; RegOperationMode(0x01) {} ; RegIrqFlags(0x12) {} ;".format(
            self.filename,
            line,
            dicta,
            dictb,
            dictc)
            )
