# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220406
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# ModBus_TCP library


import socket
import struct
from random import randint
import sys
from os import path
from time import sleep
from config import ERR_ERROR

"""
ADU     Application Data Unit = MBAP + PDU
MBAP    MODBUS Application Protocol
PDU     Protocol Data Unit

REQUEST
    MBAP    MODBUS Application Protocol
            TransactionId   Protocol        Length      Unit        Address Message
            2 Bytes         2 Bytes         2 Bytes     1 Byte      N Bytes

    PDU     Protocol Data Unit
            Offset      Length          Description             Values
            0           Byte            Function Code           03
            2           Word            First input address     0000h - FFFFh
            4           Word            Register                count 0001 - 007Dh

RESPONSE
    MBAP    MODBUS Application Protocol
            Transaction  Id Protocol    Length      Unit        Address Message
            2 Bytes     2 Bytes         2 Bytes     1 Byte      N Bytes

    PDU     Protocol Data Unit
            Offset      Length          Description             Values
            0           Byte            Function Code           03
            1           Byte            Byte Count              RegisterCount * 2
            2           N Bytes         Register data           ...

EXCEPTION_RESPONSE
    MBAP    MODBUS Application Protocol
                Transaction  Id Protocol    Length      Unit        Address Message
                2 Bytes     2 Bytes         2 Bytes     1 Byte      N Bytes

    PDU     Protocol Data Unit
            Offset      Length          Description             Values
            0           Byte            Function Code           83h
            1           Byte            Exception Code          Exception Codes

            Exception Codes                            
                                        Illegal function        01h 
                                        Illegal data address    02h
                                        Illegal data value      03h
                                        Slave device failure    04h

"""

# READ_COILS = 0x01
# READ_DISCRETE_INPUTS = 0x02
READ_HOLDING_REGISTERS = 0x03
# READ_INPUT_REGISTERS = 0x04
# WRITE_SINGLE_COIL = 0x05
# WRITE_SINGLE_REGISTER = 0x06
# WRITE_MULTIPLE_COILS = 0x0F
# WRITE_MULTIPLE_REGISTERS = 0x10
# MODBUS_ENCAPSULATED_INTERFACE = 0x2B

# ERR_ERROR = -99

class ModbusTCP:
    """
    A  Modbus TCP client, modified and reduced.
    """

    def __init__(self, host='Localhost', port=1502, unit_id=2, timeout=30.0, logger=None):
        """
        """
        self.filename = path.split(__file__)[1]
        self.logger = logger
        if not self.logger:
            print(self.filename, "ERROR No logger defined. Exit programm")
            sys.exit(0)
        self.logger.info("File: {} ; {} ; {} ; {}".format(self.filename, "ModbusTCP started", "--", "--"))

        self._sock = None  # socket
        self._transaction_id = 0  # MBAP transaction ID
        self.host = host
        self.port = port
        self.unit_id = unit_id
        self.timeout = timeout

    def is_open(self):
        """Get current status of the TCP connection (True = open)."""
        if self._sock:
            return self._sock.fileno() > 0
        else:
            return False

    def open(self):
        """Connect to modbus server (open TCP connection). """
        try:
            self._open()
            return True
        except Exception as e:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "TCP connection failed", "--", e))
            return False

    def _open(self):
        """Connect to modbus server (open TCP connection)."""
        # open an already open socket -> reset it
        if self.is_open:
            self.close()

        # set timeout as default for all sockets
        socket.setdefaulttimeout(self.timeout)
        # get available socket on target host and use the first valid connection
        for res in socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC, socket.SOCK_STREAM):
            af, sock_type, proto, canon_name, sa = res
            try:
                self._sock = socket.socket(af, sock_type, proto)
            except socket.error:
                continue
            try:
                self._sock.connect(sa)
            except socket.error:
                self._sock.close()
                continue
            break
        # check connect status
        if not self.is_open:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "Failed to open socket", "--", "--",))

    def close(self):
        """Close current TCP connection."""
        if self._sock:
            self._sock.close()
            return True

    def read_holding_registers(self, reg_addr, reg_nb=1):
        """
        Modbus function 0x03, READ_HOLDING_REGISTERS.
        Returns a data list or an error
        """

        # make request
        try:
            tx_pdu = struct.pack('>BHH', READ_HOLDING_REGISTERS, reg_addr, reg_nb)
            rx_pdu = self._req_pdu(tx_pdu=tx_pdu, rx_min_len=3)
            # extract field "byte count"
            byte_count = rx_pdu[1]
            # frame with regs value
            f_regs = rx_pdu[2:]
            # check rx_byte_count: buffer size must be consistent and have at least the requested number of registers
            if byte_count < 2 * reg_nb or byte_count != len(f_regs):
                self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Byte count mismatch", "--"))
            # allocate a reg_nb size list
            registers = [0] * reg_nb
            # fill registers list with register items
            for i in range(reg_nb):
                registers[i] = struct.unpack('>H', f_regs[i * 2:i * 2 + 2])[0]
            # return registers list
            return reg_addr, registers
        # handle error during request
        except Exception as e:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "TX", "Request error", e))
            return ERR_ERROR

    def _send(self, frame):
        """Send frame ADU = MBAP + PDU over current socket.
        """
        # check socket
        if not self.is_open:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "TX", "Try to send to a closed socket", "--"))
        # send
        try:
            self._sock.send(frame)
        except socket.timeout:
            self._sock.close()
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "TX", "Socket timeout", "--"))
        except socket.error:
            self._sock.close()
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "TX", "Socket error", "--"))

    def _send_pdu(self, pdu):
        """
        Build ADU = MBAP + PDU and send it.
        """
        # add MBAP header to PDU
        tx_frame = self._add_mbap(pdu)
        # send frame with error check
        self._send(tx_frame)
        self.logger.debug("File: {} ; {} ; {} ; {}".format(self.filename, "TX", "MBAP+PDU", tx_frame))

    def _recv(self, nbytes):
        """
        Receive nbytes of data over current socket.
        """
        r_buffer = b''
        try:
            while not r_buffer:
                r_buffer = self._sock.recv(nbytes)
        except socket.timeout:
            self._sock.close()
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Socket timeout", "--"))
        except socket.error:
            r_buffer = b''
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Socket error", "--"))
        # handle recv error
        if not r_buffer:
            self._sock.close()
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Buffer is empty", "--"))
            # r_buffer = b'ERR_ERROR'
            r_buffer = ERR_ERROR
        return r_buffer

    def _recv_all(self, nbytes):
        """
        Receive data over current socket, loop until all bytes are received (avoid TCP frag).
        """
        r_buffer = b''
        for count in range(3):
            if len(r_buffer) < nbytes:
                r_buffer += self._recv(nbytes - len(r_buffer))
                sleep(0.05)
                continue
            return r_buffer
        self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Buffer error", r_buffer))
        return ERR_ERROR

    def _recv_pdu(self, min_len=2):
        """
        Receive the modbus PDU (Protocol Data Unit) and returns PDU or None if error
        """
        # receive 7 bytes header (MBAP)
        rx_mbap = self._recv_all(7)
        # decode MBAP
        (f_transaction_id, f_protocol_id, f_length, f_unit_id) = struct.unpack('>HHHB', rx_mbap)
        # check MBAP fields
        if f_transaction_id != self._transaction_id:
            self.close()
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "MBAP checking error", rx_mbap))
        # recv PDU
        rx_pdu = self._recv_all(f_length - 1)
        # dump frame
        self.logger.debug("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "MBAP+PDU", rx_mbap+rx_pdu))
        # body decode
        # check PDU length for global minimal frame (an except frame: func code + exp code)
        if len(rx_pdu) < 2:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "PDU length <2, too short", "--"))

        # extract function code
        rx_fc = rx_pdu[0]
        # check except status
        if rx_fc >= 0x80:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", "Exception code", rx_pdu[1]))
        # check PDU length for specific request set in min_len (keep this after except checking)
        if len(rx_pdu) < min_len:
            self.logger.error("File: {} ; {} ; {} ; {}".format(self.filename, "RX", " PDU length is too short", rx_pdu))
        # if no error, return PDU
        return rx_pdu

    def _add_mbap(self, pdu):
        """Return full modbus ADU frame with MBAP + PDU.
        """
        # build MBAP
        self._transaction_id = randint(0, 65535)
        protocol_id = 0
        length = len(pdu) + 1
        mbap = struct.pack('>HHHB', self._transaction_id, protocol_id, length, self.unit_id)
        # full modbus/TCP frame = [MBAP]PDU
        return mbap + pdu

    def _req_pdu(self, tx_pdu, rx_min_len=2):
        """
        Request processing (send and recv PDU).
        rx_min_len: min length of receive PDU

        Returns the received PDU or None
        """
        self._send_pdu(tx_pdu)
        return self._recv_pdu(min_len=rx_min_len)
