# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# Read Solar Plant Data


from modbustcp import ModbusTCP
from tools import ModbusTools
import config as cfg
from os import path
from datetime import datetime


class SolarPlantRegisters:
    """
    A class to read the Solar Edge inverter registers via the modbus-tcp protocol.
    """
    def __init__(self, logger=None):

        self.filename = path.split(__file__)[1]
        self.mbustools = ModbusTools()

        self.logger = logger
        self.logger.info("File: {} ; {} ; {} ; {}".format(
                self.filename,
                "Read solar plant registers started.....", "--", "--"))
        self.reg_update_time = datetime.now

        self.reg_inv_common_block = []    # inverter 1
        self.reg_inv_block1 = []  # inverter 1
        self.reg_m_common_block = []  # meter
        self.reg_met_block1 = []  # meter
        self.reg_bat_common_block = []  # battery
        self.reg_bat_block1 = []  # battery
        self.reg_bat_related_energy = []  # battery
        self.reg_i_status = []

        self.mbus = ModbusTCP(
            host=cfg.ModBusTCP['tcp_url'],
            port=cfg.ModBusTCP['tcp_port'],
            # unit_id=cfg.SolarPlantUnit['PowerMeter']['mb_id'],
            unit_id=cfg.ModBusTCP['unit_id'],
            timeout=cfg.ModBusTCP['mb_timeout'],
            logger=self.logger
            )

    def read_holding_regs(self):
        """
        Opens the socket, read some registers and closes it.
        SolareEdge recommends to read registers in blocks instead consecutive single register
        heater_only = True:  Reads and processes only registers used by the heater control  module
        Returns the registers block list if successful or cf.ERR_ERROR
        """
        if self.mbus.open():
            """
            Read a block of modbus registers as a tuple
            (reg-address,[regx, regx+1, regx+2, ....  ])
            """
            returnvalue = 0
            try:
                self.reg_inv_common_block = self.mbus.read_holding_registers(40004, 65)  # inverter 1
                self.reg_inv_block1 = self.mbus.read_holding_registers(40072, 30)  # inverter 1
                self.reg_m_common_block = self.mbus.read_holding_registers(40123, 66)   # meter
                self.reg_met_block1 = self.mbus.read_holding_registers(40190, 60)   # meter
                self.reg_bat_common_block = self.mbus.read_holding_registers(57600, 51)   # battery
                self.reg_bat_block1 = self.mbus.read_holding_registers(57712, 24)   # battery
                self.reg_bat_related_energy = self.mbus.read_holding_registers(57666, 2)  # battery
                self.reg_i_status = self.mbus.read_holding_registers(40107, 1)[1]

                if (self.reg_inv_common_block or self.reg_inv_block1 or self.reg_m_common_block or
                    self.reg_met_block1 or self.reg_bat_common_block or self.reg_bat_block1 or
                    self.reg_bat_related_energy or self.reg_i_status) == cfg.ERR_ERROR:
                    returnvalue = cfg.ERR_ERROR
            except Exception as e:
                self.logger.error("File: {} ; {} ; Register read error {} ;  {} ; {}".format(self.filename, "--", e,  "--", "--"))
            finally:
                self.mbus.close()
                self.reg_update_time = datetime.now
                return returnvalue

    def inv_block1_scale16(self, reglist_key="40079X"):
        """
        Returns the values of registers in this specific register block1.
        The register start address and the number of registers to read are defined
        in the register list with key <reglist_key>.
        The values are calculated with the scalefactor methode.
        """
        regdict = cfg.RegsInverter
        reglist = regdict[reglist_key]
        reg_block = self.reg_inv_block1
        data = self.sfu_calc_with_scalefactor(reg_block, regdict, reglist)
        return reglist[0], reglist[10], reglist[8], data  # Data source, Disp text, Display unit, Data

    def met_block1_scale16(self, reglist_key="40196X"):
        """
        Returns the values of registers in this specific register block1.
        The register start address and the number of registers to read are defined
        in the register list with key <reglist_key>.
        The values are calculated with the scalefactor methode.
        """
        regdict = cfg.RegsMeter
        reglist = regdict[reglist_key]
        reg_block = self.reg_met_block1
        data = self.sfu_calc_with_scalefactor(reg_block, regdict, reglist)
        return reglist[0], reglist[10], reglist[8], data  # Data source, Disp text, Display unit, Data

    def meter_info(self):
        """
        Reads a block of registers and does a direct interpretation.
        No need for a  register definition in the config file.
        """
        reg_block = self.reg_m_common_block[1]  # skip register addres
        m_c_manufacturer = reg_block[0:16]  # self.reg_m_common_block[0:16]
        m_c_model = reg_block[16:32]
        m_c_version = reg_block[40:48]
        m_c_serial_number = reg_block[48:64]
        m_c_device_address = reg_block[64:65]
        m_c_meter_type = reg_block[65:66]
        m_man = self.mbustools.intlist_to_text(m_c_manufacturer)
        m_sn = self.mbustools.intlist_to_text(m_c_serial_number)
        meter = "--Meter-- {} Type: {} (203=Wye-Connect Three Phase)   Modbus ID: {}  Serial Number: {}"\
                .format(m_man, m_c_meter_type, m_c_device_address, m_sn)
        return meter

    def inverter_master_info(self):
        reg_block = self.reg_inv_common_block[1]  # skip register address
        im_c_manufacturer = self.reg_m_common_block[0:16]
        im_c_model = reg_block[16:32]
        im_unknown = reg_block[32:42]   # unknown
        im_c_version = reg_block[42:50]
        im_c_serial_number = reg_block[50:66]
        im_c_device_address = reg_block[66:67]
        im_sn = self.mbustools.intlist_to_text(im_c_serial_number)
        inverter_master = "--Inverter 1 -- Modbus ID: {}  Serial Number: {}".format(im_c_device_address, im_sn)
        return inverter_master

    def battery_info(self):
        reg_block = self.reg_bat_common_block[1]  # skip register addres
        bat_c_manufacturer = reg_block[0:16]
        bat_c_model = reg_block[16:32]
        bat_c_version = reg_block[32:48]
        bat_c_serial_number = reg_block[48:64]
        bat_c_device_address = reg_block[64:65]
        bat_man = self.mbustools.intlist_to_text(bat_c_manufacturer)
        bat_sn = self.mbustools.intlist_to_text(bat_c_serial_number)
        bat_model = self.mbustools.intlist_to_text(bat_c_model)
        battery = "--Battery-- {} Modbus ID: {}  Serial Number: {}  Model{}"\
            .format(bat_man, bat_c_device_address, bat_sn, bat_model)
        return battery

    def bat_related_energy(self, reglist=cfg.RegsBattery["57666"]):
        """Reads the content of 2 registers and converts it to a float32 number"""
        reg_block = self.reg_bat_related_energy
        reg = self.read_from_regblock(reg_block, reglist)
        data = self.mbustools.doublereg_to_float(reg[1], reg[0])   # order is big endian,
        values = [round(d * reglist[7], reglist[9]) for d in data]  # unitfactor = reglist[7], roundvalue = reglist[9]
        return reglist[0], reglist[10], reglist[8], values  # Reg. address, hort name, Disp text, Display unit, data

    def bat_inst_power(self, reglist=cfg.RegsBattery["57716"]):
        """Reads the content of 2 registers and converts it to a float32 number"""
        reg_block = self.reg_bat_block1
        reg = self.read_from_regblock(reg_block, reglist)
        data = self.mbustools.doublereg_to_float(reg[1], reg[0])   # order is big endian,
        values = [round(d * reglist[7], reglist[9]) for d in data]  # unitfactor = reglist[7], roundvalue = reglist[9]
        return reglist[0], reglist[10], reglist[8], values  # Reg. address, hort name, Disp text, Display unit, data

    def bat_available_energy(self, reglist=cfg.RegsBattery["57728"]):
        """Reads the content of 2 registers and converts it to a float32 number"""
        reg_block = self.reg_bat_block1
        reg = self.read_from_regblock(reg_block, reglist)
        data = self.mbustools.doublereg_to_float(reg[1], reg[0])   # order is big endian,
        values = [round(d * reglist[7], reglist[9]) for d in data]  # unitfactor = reglist[7], roundvalue = reglist[9]
        return reglist[0], reglist[10], reglist[8], values  # Reg. address, hort name, Disp text, Display unit, data

    def bat_state_of_health(self, reglist=cfg.RegsBattery["57730"]):
        """Reads the content of 2 registers and converts it to a float32 number"""
        reg_block = self.reg_bat_block1
        reg = self.read_from_regblock(reg_block, reglist)
        data = self.mbustools.doublereg_to_float(reg[1], reg[0])   # order is big endian,
        values = [round(d * reglist[7], reglist[9]) for d in data]  # unitfactor = reglist[7], roundvalue = reglist[9]
        return reglist[0], reglist[10], reglist[8], values  # Reg. address, hort name, Disp text, Display unit, data

    def bat_state_of_energy(self, reglist=cfg.RegsBattery["57732"]):
        """Reads the content of 2 registers and converts it to a float32 number"""
        reg_block = self.reg_bat_block1
        reg = self.read_from_regblock(reg_block, reglist)
        data = self.mbustools.doublereg_to_float(reg[1], reg[0])   # order is big endian,
        values = [round(d * reglist[7], reglist[9]) for d in data]  # unitfactor = reglist[7], roundvalue = reglist[9]
        return reglist[0], reglist[10], reglist[8], values  # Reg. address, short name, Disp text, Display unit, data

    def m_energy(self, reglist_key="40226"):
        """Reads the content of 2 registers and converts it to an uint32 number"""
        regdict = cfg.RegsMeter
        reglist = regdict[reglist_key]
        reg_block = self.reg_met_block1  #
        reg_data = self.read_from_regblock(reg_block, reglist)  # read register value from modbus register block
        # combine two 16bit register data to a 32 bit (uint32) numbers and make a one element list
        doublereg_value = [self.mbustools.doublereg_to_uint32(reg_data[0], reg_data[1])]
        scalefactor = self.read_from_regblock(reg_block, regdict[reglist[4]])  # scalefactor key
        data = self.calc_scalefactor_values(reglist[0], doublereg_value, scalefactor[0], unitfactor=reglist[7])
        return reglist[1], reglist[10], reglist[8], data

    def i_status(self):
        name = 'I_Status'
        description = 'Inverter Operating State)'
        status = self.reg_i_status
        data = cfg.InverterStatus[str(status[0])]
        return name, description, status, data

    def m_ac_freq(self, reglist_key="40204"):
        """
        Example for reading a specific register.
        Could also be done with the block methode.
        """
        return self.met_block1_scale16(reglist_key=reglist_key)

    def i_dc_voltage(self, regdict=cfg.RegsInverter, reglist_key="40098"):
        """
         Example for reading a specific register.
         Could also be done with the block methode.
         """
        reglist = regdict[reglist_key]
        reg_block = self.reg_inv_block1
        data = self.sfu_calc_with_scalefactor(reg_block, regdict, reglist)
        return reglist[1], reglist[10], reglist[8], data  # Short-Name, Disp text, Display unit, Data

    def i_dc_current(self, regdict=cfg.RegsInverter, reglist_key="40096"):
        """
         Example for reading a specific register.
         Could also be done with the block methode.
         """
        reglist = regdict[reglist_key]
        reg_block = self.reg_inv_block1
        data = self.sfu_calc_with_scalefactor(reg_block, regdict, reglist)
        return reglist[1], reglist[10], reglist[8], data  # Short-Name, Disp text, Display unit, Data

    def read_from_regblock(self, regblock, startreg):
        """ Reads a chunk of consecutive registers from the register block """
        addr, block = regblock
        read_start = startreg[1] - addr
        read_end = read_start + startreg[2]
        reg = block[read_start: read_end]
        return reg

    def sfu_calc_with_scalefactor(self, reg_block, regdict, reglist):
        """
        Takes a register block and a register address and returns
        the specific register value, calculated with the scalefactor methode.
        """
        reg = self.read_from_regblock(reg_block, reglist)  # read register value from modbus register block
        cfg_scalefactor = regdict[reglist[4]]  # scalefactor key
        scalefactor = self.read_from_regblock(reg_block, cfg_scalefactor)
        data = self.calc_scalefactor_values(reglist[0], reg, scalefactor[0], reg_type=reglist[3])
        return data

    def calc_scalefactor_values(self, name, reg_data, scalefactor, reg_type=cfg.DT_INT16, unitfactor=1, roundvalue=2):
        """
        Uses the scalefactor methode to calculate the value 16bit register
        reg_data must be a list. Enclose simple variables in brackets.
        valid reg_types:            DT_INT16,DT_UINT16, DT_SUNSFF
        scalefactor muast be a single number
        valid scalefactor_type:     DT_INT16
        """
        #values = None
        if type(reg_data) is not list:
            self.logger.error(
                "File: {} ; {} ; Type error {} ; Register {} ; {}"
                .format(self.filename, name, "Register type must be a list", reg_data, "--"))

        if reg_type == cfg.DT_INT16 or reg_type == cfg.DT_SUNSSF:
            rd = [self.mbustools.sunssf_to_decimal(r) for r in reg_data]
        elif reg_type == cfg.DT_UINT16:
            rd = reg_data

        else:
            self.logger.error(
                "File: {} ; {} ; Reg_Data {} ; Unknown Register type {} ; {}"
                .format(self.filename, name, reg_data, reg_type, "--"))

        sf = self.mbustools.sunssf_to_decimal(scalefactor)
        if -10 < sf < 10:  # valid range of sunspec factors
            values = [round(v * pow(10, sf) * unitfactor, roundvalue) for v in rd]
            return values
        else:
            self.logger.error(
                "File: {} ; {} ; Reg_Data {} ; Reg_SF {} ; Scalefactor error, valid range: -10 ..+10 {}"
                .format(self.filename, name, reg_data, scalefactor, sf))
            return [0,0]

    def get_heater_power(self):
        """
        Reads the necessary register data to control the heater and
        returns the value in Watts
        """
        # if self.read_holding_regs(heater_only=True) == cfg.ERR_ERROR:
        #     self.logger.error("File: {} ; {} ; ModBusTCP read error {} ;  {} ; {}".format(self.filename, "--", "--", "--",  "--"))
        #     return cfg.ERR_ERROR
        return self.met_block1_scale16(reglist_key="40206")[3][0]

    def i_dc_ac_conversion_loss(self, i_dc_power, i_ac_power):
        """ Calculates the invertes DC-AC conversion loss """
        try:
            i_dc_ac_conversion_loss = round((i_dc_power - i_ac_power) / i_dc_power, 2)
        except ZeroDivisionError:
            i_dc_ac_conversion_loss = 0
            self.logger.warning("File: {} ; {}; {}; {}; {}".format(
                self.filename, "DC Power produced", "Set DC-AC conversion loss to 0%", "--", "--"))
        return i_dc_ac_conversion_loss

    def archive(self):
        exported_energy = self.m_energy(reglist_key="40226")[3][0]  # Exported Real Energy
        imported_energy = self.m_energy(reglist_key="40234")[3][0]  # Imported Real Energy
        battery_energy = self.bat_state_of_energy()[3][0] # Energy in %

        txt_first = "{};{};".format(datetime.now(), "pv")
        txt1 = "{};{};{};{}".format("reg_40226", exported_energy, "kWh", "Exported Energy")
        txt2 = "{};{};{};{}".format("reg_40234", imported_energy, "kWh", "Imported Energy")
        txt3 = "{};{};{};{}".format("reg_57666", battery_energy, "%", "Battery_energy")

        with open(cfg.ARCHIVE_FILE, "a") as f:
            f.writelines(txt_first + txt1 + ";--\n")
            f.writelines(txt_first + txt2 + ";--\n")
            f.writelines(txt_first + txt3 + ";--\n")
        self.logger.debug("File: {}; {}; {}; {}; {}; {}".format(self.filename, cfg.ARCHIVE_FILE, txt_first, txt1, txt2, txt3))

if __name__ == "__main__":
    exec(open("solarplant_run.py").read())
