# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609 / V13
#   Project:                50 Solar
# ------------------------------------------------------------------------------
#
# Solar Plant tools
#


import config as cfg
import pigpio
from time import sleep
from raspilora import LoRa
from os import path
from struct import unpack
from oled_ssd1306 import OledI2c
import logging
from logging.handlers import RotatingFileHandler


class SolarTools:
    def __init__(self):

        self.filename = path.split(__file__)[1]
        self.pi = None
        self.i2c = None

    def init_logger(self,
                    logconsole=True,
                    logfile=None,
                    loglevel=10,
                    filesize=1024,
                    filebackups=1):

        """
        Init logging and print log to file and to console ( if logconsole=True )
        """
        # Gets or creates a logger
        logger = logging.getLogger(__name__)
        # set log level
        logger.setLevel(loglevel)
        fmt = logging.Formatter(
                        '%(asctime)s ;'
                        ' %(levelname)s ;'
                        'Function: %(funcName)s ; '
                        'Line: %(lineno)d ; '
                        '%(message)s'
                        )
        if logfile:
            # file handler
            fh = RotatingFileHandler(logfile, mode='a', encoding='utf-8', maxBytes=filesize, backupCount=filebackups)
            fh.setFormatter(fmt)
            logger.addHandler(fh)
        if logconsole:
            # console handler
            ch = logging.StreamHandler()
            ch.setFormatter(fmt)
            logger.addHandler(ch)
        return logger

    def init_pigpio_spi_i2c(self, addr, port):
        """
            Connect to the pigpiod daemon and initialize  the SPI  bus
            Don't forget to start the pigpiod daemon on the raspi:
              sudo pigpiod  -p 25000      ( remote)
              sudo pigpiod                        ( local)
        """
        #  Connect the remote machine to the raspi pigpiod daemon
        text = "Pigpio not connected"

        self.pi = pigpio.pi(addr, port)
        if self.pi.connected:
            text = "Pigpio connected ...."

        spi = self.pi.spi_open(cfg.Spi["Bus"], cfg.Spi["Baudrate"], cfg.Spi["Mode"])
        self.i2c = self.pi.i2c_open(cfg.I2c["Bus"], cfg.OLed["I2cAddr"])
        return self.pi, spi, self.i2c, self.filename, text


class OledTools:
    def __init__(self, pi, i2c):
        self.pi = pi
        self.i2c = i2c
        self.disp = OledI2c(self.pi, self.i2c, cfg.OLed["PixWidth"], cfg.OLed["PixHeight"])

    def oled_init(self):
        self.disp.init_display()
        self.disp.disp_on(True)
        self.disp.clear(0)

    def oled_pixelimage(self, pic, xoffset, yoffset):
        self.disp.pixelimage(pic, xoffset, yoffset)

    def oled_startup_screen(self):
        self.disp.clear(0)
        self.disp.text(cfg.Oled_StartScreen["TextA"], 20, 32)
        self.disp.show()
        sleep(3)
        self.disp.clear(0)
        self.disp.text(cfg.Oled_StartScreen["TextB"], 20, 20)
        self.disp.text(cfg.Oled_StartScreen["TextC"], 20, 40)
        self.disp.show()
        sleep(5)
        self.disp.clear(0)

        with open(cfg.Oled_StartScreen["Logo"], 'r') as f:
            pic = f.readlines()
        self.oled_pixelimage(pic, 30, 3)
        self.disp.show()
        sleep(5)

    def oled_text(self, text, x, y):
        self.disp.text(text, x, y)

    def oled_scroll(self, stack, text):
        self.disp.clear(0)
        for i in range(3):
            stack[i] = stack[i + 1]
        stack.pop()
        stack.append(text)
        for i, s in enumerate(stack):
            self.disp.text(s, 0, i * 12)
        self.disp.show()


class ModbusTools:

    def __init__(self):
        pass

    def sunssf_to_decimal(self, integer):
        """
        Converts a 16bit unsigned  number  to a signed number.
        E.g. 65534 --> -2
        """
        if integer > 0x8000:
            return -(0xFFFF - integer + 1)
        else:
            return integer

    def intlist_to_text(self, intlist):
        """ Splits each 16bit integer of the list into bytes and returns the ascii characters"""
        text = ""
        for i in intlist:
            if i == 0:
                continue
            hi = chr(i >> 8)
            lo = chr(i & 0x00ff)
            if not hi.isprintable():
                hi = ""
            if not lo.isprintable():
                lo = ""
            text = text + hi + lo
        return [text]

    def doublereg_to_uint32(self, reghigh, reglow):
        """ Combines two 16 bit registers to a 32 bit uint """
        return (reghigh << 16) ^ reglow

    def four_regs_to_uint64(self, regfour, regthree, regtwo, regone):   # big order
        """ Combines four 16 bit registers to a 64 bit uint """
        return (regfour << 48) ^ (regthree << 32) ^ (regtwo << 16) ^ regone

    def doublereg_to_float(self, reghigh, reglow):
        """
        Combines two 16 bit registers to a 32 bit number and
        returns an IEEE 754 signed float value.
        """
        reghigh = reghigh.to_bytes(2, byteorder='big', signed=False)
        reglow = reglow.to_bytes(2, byteorder='big', signed=False)
        return unpack('!f', reghigh + reglow)


class LoRaTools:
    def __init__(self, pi, spi, logger):
        self.filename = path.split(__file__)[1]
        self.logger = logger
        self.logger.info("File: {} ; {} ; {} ; {}".format(self.filename, "LoRa tools started.....", "--", "--"))
        self.pi = pi
        self.spi = spi
        self.init_pins()
        self.lora_reset()
        self.lora = LoRa(self.spi,  pi=self.pi, logger=self.logger)
        self.irq_dio0 = None
        self.show_modem_config()

    def init_pins(self):
        self.pi.set_mode(cfg.LoraPins["NSS"], pigpio.OUTPUT)
        self.pi.set_mode(cfg.LoraPins["DIO0"], pigpio.INPUT)
        self.pi.set_pull_up_down(cfg.LoraPins["DIO0"], pigpio.PUD_DOWN)
        self.pi.set_mode(cfg.LoraPins["Reset"], pigpio.OUTPUT)
        self.pi.set_pull_up_down(cfg.LoraPins["Reset"], pigpio.PUD_UP)

    def get_lorapins(self):
        return cfg.LoraPins["Reset"], cfg.LoraPins["NSS"], cfg.LoraPins["DIO0"]

    def lora_reset(self):
        """ Reset LoRa device, do a reset pulse """
        self.pi.write(cfg.LoraPins["Reset"], 1)
        sleep(.01)
        self.pi.write(cfg.LoraPins["Reset"], 0)
        sleep(.01)
        self.pi.write(cfg.LoraPins["Reset"], 1)
        sleep(0.1)

    def cancel_recv(self, ):
        """ Cancel  pin  interrrupt """
        self. irq_dio0.cancel()

    def init_recv_mode(self, callback):
        """
        Start the Raspi /ESP/Pico  Rx  interrupt and wait for incoming messages .
        callback is a handle, pointing to the function that will be processed when an interrupt occurs.
        """
        self.irq_dio0 = self.pi.callback(cfg.LoraPins["DIO0"], pigpio.RISING_EDGE, callback)
        self.lora.set_rx_mode()

    def send(self, tx_data):
        self.lora.send(tx_data)

    def set_recv_mode(self):
        self.lora.set_rx_mode()

    def get_recv_data(self):
        return self.lora.get_rx_data()

    def get_packet_rssi_strength(self):
        return self.lora.get_packet_strength()

    def get_irq(self):
        return self.irq_dio0

    def pi_stop(self):
        self.pi.stop()

    def show_register(self, line_no=00, register=0x01, text="--"):
        """ line_no:  line number of the python script"""
        xx = self.lora.read_register(register)
        modemconfig = "0x{:02x}\t{}\t0x{:02x}\t{:08b}".format(register, xx, xx, xx)  # decimal, hex, binary
        self.logger.debug("File: {} ; {} ; {} ; {}".format(self.filename, "LoRa modem config data", modemconfig, "--"))

    def read_modem_status(self, short=True):
        ms = self.lora.read_modem_status()
        if short:
            return "0x{:X} {:08b}".format(ms, ms)
        return dict(
            rx_coding_rate=ms >> 5 & 0x03,
            modem_clear=ms >> 4 & 0x01,
            header_info_valid=ms >> 3 & 0x01,
            rx_ongoing=ms >> 2 & 0x01,
            signal_sync=ms >> 1 & 0x01,
            signal_detected=ms >> 0 & 0x01
        )

    def read_irq_flags(self, short=True):
        iqf = self.lora.read_irq_flags()
        if short:
            return "0x{:X} {:08b}".format(iqf, iqf)
        return dict(
            rx_timeout=iqf >> 7 & 0x01,
            rx_done=iqf >> 6 & 0x01,
            crc_error=iqf >> 5 & 0x01,
            iqvalid_header=iqf >> 4 & 0x01,
            tx_done=iqf >> 3 & 0x01,
            cad_done=iqf >> 2 & 0x01,
            fhss_change_ch=iqf >> 1 & 0x01,
            cad_detected=iqf >> 0 & 0x01,
        )

    def read_operation_mode(self, short=True):
        om = self.lora.read_operation_mode()
        if short:
            return "0x{:X} {:08b}".format(om, om)
        return dict(
            LongRangeMode=om >> 7 & 0x01,      # 7
            AccessSharedReg=om >> 6 & 0x01,    # 6
            reserved=om >> 4 & 0x03,   # 5-4
            LowFrequencyModeOn=om >> 3 & 0x01,     # 3
            mode=om >> 0 & 0x07,       # 2-0
        )

    def debug(self, line, irq_dio0=None):
        print(line, self.read_operation_mode())
        print(line, self.read_irq_flags())
        print(line, self.read_modem_status())

    def show_modem_config(self):
        self.show_register(line_no=00, register=0x1d, text="REG_MODEM_CONFIG_1")
        self.show_register(line_no=00, register=0x1e, text="REG_MODEM_CONFIG_2")
        self.show_register(line_no=00, register=0x26, text="REG_MODEM_CONFIG_3")


class Misc:
    """
    pigpio.tickDiff(t1, t2)
        Returns the microsecond difference between two ticks.

        Parameters

        t1:= the earlier tick
        t2:= the later tick
    """
    def sort_dict(self, dict_unsorted):
        """
        Sorts a dictionary by values ( numbers or strings) and return the
        sorted dictionary and the highest value
        """
        dict_sorted = dict(sorted(dict_unsorted.items(), key=lambda item: item[1], reverse=True))
        # get first item / highest value
        for item in dict_sorted.items():
            value_high = item[1]
            break
        return dict_sorted, value_high
