# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609 / V13
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# Lora SX1276 Micropython driver for ESP32 and Raspi-Pico RP2


from time import ticks_ms, ticks_diff
import gc
import config as cfg

TX_BASE_ADDR = 0x00
RX_BASE_ADDR = 0x00

PA_BOOST = 0x80
PA_OUTPUT_RFO_PIN = 0
PA_OUTPUT_PA_BOOST_PIN = 1

REG_FIFO = 0x00
REG_OP_MODE = 0x01
REG_FRF_MSB = 0x06
REG_FRF_MID = 0x07
REG_FRF_LSB = 0x08
REG_PA_CONFIG = 0x09
REG_LNA = 0x0c
REG_FIFO_ADDR_PTR = 0x0d
REG_FIFO_TX_BASE_ADDR = 0x0e
REG_FIFO_RX_BASE_ADDR = 0x0f
REG_FIFO_RX_CURRENT_ADDR = 0x10
REG_IRQ_FLAGS = 0x12
REG_RX_NB_BYTES = 0x13
REG_MODEM_STAT = 0x18
REG_PKT_SNR_VALUE = 0x19
REG_PKT_RSSI_VALUE = 0x1a
REG_RSSI_VALUE = 0x1b
REG_MODEM_CONFIG_1 = 0x1d
REG_MODEM_CONFIG_2 = 0x1e
REG_PREAMBLE_MSB = 0x20
REG_PREAMBLE_LSB = 0x21
REG_PAYLOAD_LENGTH = 0x22
REG_MODEM_CONFIG_3 = 0x26
REG_DETECTION_OPTIMIZE = 0x31
REG_DETECTION_THRESHOLD = 0x37
REG_SYNC_WORD = 0x39
REG_DIO_MAPPING_1 = 0x40
REG_VERSION = 0x42

MODE_LORA = 0x80    # Bit LOWFREQUENCYEMODEON = 0 -> High Frequency Mode ( > 779MHz) = on
MODE_SLEEP = 0x00
MODE_STDBY = 0x01
MODE_TX = 0x03
MODE_RX_CONTINUOUS = 0x05

IRQ_TX_DONE_MASK = 0x08
IRQ_RX_DONE_MASK = 0x40
IRQ_PAYLOAD_CRC_ERROR_MASK = 0x20
MAX_PKT_LENGTH = 255

ERR_ERROR = cfg.ERR_ERROR  # F40Solar. general  error tag


class LoRa:
    def __init__(self, spi, lora_pins):
        self.spi = spi
        self.rst, self.nns, self.rxirq = lora_pins
        self.check_spi_connection()
        self.init_lora()
        self.standby()

    def check_spi_connection(self):
        version = self._read(REG_VERSION)
        if version == 0x00 or version == 0xff:
            print("Warning: Unknown LoRa chip version or bad SPI connection", hex(version))
        else:
            print("LoRa SPI connected")

    def init_lora(self):
        self.lora_sleep()
        self.set_frequency()

        self.set_preamble_length(cfg.LoRaRadio["PreambleLength"])
        # set LNA boost
        self._write(REG_LNA, self._read(REG_LNA) | 0x03)
        # set auto AGC

        self.set_tx_power(cfg.LoRaRadio["TxPowerLevel"])
        self._write(REG_MODEM_CONFIG_1, cfg.LoRaRadio["RegConfig1"])
        self._write(REG_MODEM_CONFIG_2, cfg.LoRaRadio["RegConfig2"])
        self._write(REG_MODEM_CONFIG_3, cfg.LoRaRadio["RegConfig3"])

        self.set_sync_word(cfg.LoRaRadio["SyncWord"])
        self._write(REG_FIFO_TX_BASE_ADDR, TX_BASE_ADDR)
        self._write(REG_FIFO_RX_BASE_ADDR, RX_BASE_ADDR)

    def set_tx_mode(self):
        """
        Sets tx mode, waits until the transmission is done and leaves
        the function
        """
        # set TX mode
        self._write(REG_OP_MODE, (MODE_LORA | MODE_TX))
        # wait until TX is done
        start = ticks_ms()
        while (self._read(REG_IRQ_FLAGS) & IRQ_TX_DONE_MASK) == 0:
            if ticks_diff(ticks_ms(), start) > 500:
                break

        # clear IRQs
        self._write(REG_IRQ_FLAGS, IRQ_TX_DONE_MASK)
        # print("lora 109", gc.mem_free())
        gc.collect()

    def send(self, data):
        """
        Sets the FIFO pointer to the base address, writes the data bytes
        into the FIFO buffer and activates the transmission.
        """
        if isinstance(data, str):
            data = data.encode()  # string to bytes
        lendata = len(data)
        self.standby()
        # reset FIFO address and payload length
        self._write(REG_FIFO_ADDR_PTR, TX_BASE_ADDR)
        self._write(REG_PAYLOAD_LENGTH, 0x1)

        for i in range(lendata):
            self._write(REG_FIFO, data[i])
        self._write(REG_PAYLOAD_LENGTH, lendata)
        # to pass to the FSTX state and wait for about a hundred microseconds;
        self.set_tx_mode()

    def _get_irq_flags(self):
        """
        Read and write back clears the rxdone bit and sets the DIO0 Pin low
        """
        reg_irq = self._read(REG_IRQ_FLAGS)
        self._write(REG_IRQ_FLAGS, reg_irq)
        return reg_irq

    def get_packet_strength(self):
        """
        Packet Strength (dBm) = -157 + PacketRssi   for >= 779.0 MHz (High Frequency)
        Packet Strength (dBm) = -157 + PacketRssi + PacketSnr * 0.25 (High Frequency and SNR < 0)
        """

        packet_rssi_dbm = self._read(REG_PKT_RSSI_VALUE) - 157
        if cfg.LoRaRadio["Freq"][0] < 779.0:
            packet_rssi_dbm = self._read(REG_PKT_RSSI_VALUE) - 164

        packet_snr = self._read(REG_PKT_SNR_VALUE)
        if packet_snr < 0:
            return packet_rssi_dbm + packet_snr * 0.25
        return packet_rssi_dbm

    def standby(self):
        self._write(REG_OP_MODE, MODE_LORA | MODE_STDBY)

    def lora_sleep(self):
        self._write(REG_OP_MODE, MODE_LORA | MODE_SLEEP)

    def set_tx_power(self, level, outputPin=PA_OUTPUT_PA_BOOST_PIN):
        if outputPin == PA_OUTPUT_RFO_PIN:
            level = min(max(level, 0), 14)
            self._write(REG_PA_CONFIG, 0x70 | level)
        else:
            level = min(max(level, 2), 17)
            self._write(REG_PA_CONFIG, PA_BOOST | (level - 2))

    def set_frequency(self):
        """
        The frf value is stored in 3 bytes
        Calculation:  see datasheets
        """
        self._write(REG_FRF_MSB, cfg.LoRaRadio["Freq"][1])
        self._write(REG_FRF_MID, cfg.LoRaRadio["Freq"][2])
        self._write(REG_FRF_LSB, cfg.LoRaRadio["Freq"][3])

    def set_implicit_header(self):
        if self._read(REG_MODEM_CONFIG_1) & 0x01 == 0x01:
            implicit = True
        else:
            implicit = False
        return implicit

    def set_spreading_factor(self):
        # Spreading factor  SF= 6
        if self._read(REG_MODEM_CONFIG_2) & 0xF == 0x96:
            self._write(REG_DETECTION_OPTIMIZE, 0xC5)
            self._write(REG_DETECTION_THRESHOLD, 0x0C)
        # Spreading factor  SF= 7 .. 12
        else:
            self._write(REG_DETECTION_OPTIMIZE, 0xC3)
            self._write(REG_DETECTION_THRESHOLD, 0x0A)

    def set_preamble_length(self, n):
        self._write(REG_PREAMBLE_MSB, (n >> 8) & 0xff)
        self._write(REG_PREAMBLE_LSB, (n >> 0) & 0xff)

    def set_sync_word(self, sw):
        self._write(REG_SYNC_WORD, sw)

    def get_rx_data(self):
        irqflags = self._get_irq_flags()
        if irqflags & IRQ_PAYLOAD_CRC_ERROR_MASK == 0:  # bit 5, PayloadCrcError:  0 = ok
            return self._read_payload()
        # print("lora 202  mem_free ", gc.mem_free())
        gc.collect()
        return ERR_ERROR      # no use, error handling not implemented

    def set_rx_mode(self):
        self._write(0x12, 0x50)
        self._get_irq_flags()
        self._write(REG_OP_MODE, MODE_LORA | MODE_RX_CONTINUOUS)

    def _read_payload(self):
        self._write(REG_FIFO_ADDR_PTR, self._read(REG_FIFO_RX_CURRENT_ADDR))
        if self. set_implicit_header():
            n = self._read(REG_PAYLOAD_LENGTH)
        else:
            n = self._read(REG_RX_NB_BYTES)
        payload = bytearray()
        for i in range(n):
            payload.append(self._read(REG_FIFO))
        gc.collect()

        try:
            pload = payload.decode('utf-8')  # bytearray to string
        except:
            pload = ERR_ERROR
        return pload

    def _transfer(self, reg_address, data_send=0x00):
        data_recv = bytearray(1)
        self.nns.value(0)
        self.spi.write(bytes([reg_address]))
        self.spi.write_readinto(bytes([data_send]), data_recv)
        self.nns.value(1)
        return data_recv

    def _read(self, addr):
        x = self._transfer(addr & 0x7f)
        return int.from_bytes(x, 'big')

    def _write(self, reg_address, data_send):
        self._transfer(reg_address | 0x80, data_send)   # bit 8 = 1 --> write

    def read_register(self, register):
        return self._read(register)

    def read_modem_status(self):
        return self._read(REG_MODEM_STAT)

    def read_irq_flags(self):
        return self._read(REG_IRQ_FLAGS)

    def read_operation_mode(self):
        return self._read(REG_OP_MODE)
