# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# ESPTTGO OLED Display Library 



import framebuf as fbuff
import font454

# register definitions
SET_CONTRAST        = 0x81   # 10.1.7 Set Contrast Control for BANK0 (81h)
SET_ENTIRE_ON       = 0xA4   # 10.1.9 Entire Display ON (A4h/A5h)
SET_DISP_NORMAL     = 0xA6   # 10.1.10 Set Normal/Inverse Display (A6h/A7h)
SET_DISP_INVERS     = 0xA7   # 10.1.10 Set Normal/Inverse Display (A6h/A7h)
SET_DISP_OFF        = 0xAE   # 10.1.12 Set Display ON/OFF (AEh/AFh)
SET_DISP_ON         = 0xAF   # 10.1.12 Set Display ON/OFF (AEh/AFh)
SET_MEM_ADDR_MODE   = 0x20   # 10.1.3 Set Memory Addressing Mode (20h)
SET_COL_ADDR        = 0x21
SET_PAGE_ADDR       = 0x22
SET_DISP_START_LINE = 0x40   # 10.1.6 Set Display Start Line (40h~7Fh)
SET_SEG_REMAP_A0    = 0xA0   # 10.1.8 Set Segment Re-map (A0h/A1h)
SET_SEG_REMAP_A1    = 0xA1   # 10.1.8 Set Segment Re-map (A0h/A1h)
SET_MUX_RATIO       = 0xA8
SET_COM_OUT_DIR     = 0xC8   # 10.1.14 Set COM Output Scan Direction (C0h/C8h)
SET_DISP_OFFSET     = 0xD3   # 10.1.15 Set Display Offset (D3h)
SET_COM_PIN_CFG     = 0xDA
SET_DISP_CLK_DIV    = 0xD5
SET_PRECHARGE       = 0xD9
SET_VCOM_DESEL      = 0xDB
SET_CHARGE_PUMP     = 0x8D


class OledI2c:
    """
    A minimized OLed SSD1306 driver version .
    For a fully functional version see Project p43
    """
    def __init__(self, i2c, address, width, height):
        self.i2c = i2c
        self.address = address
        self.width = width
        self.height = height
        self.pages = self.height // 8
        self.init_display()
        self.disp_on(True)

        buffer = bytearray(width * self.pages)
        self.fb = fbuff.FrameBuffer(buffer, width, height, fbuff.MVLSB)
        self.buffer = buffer

    # i2c ---------
    def write_cmd(self, cmd):
        self.i2c.writeto_mem(self.address, 0x00, bytearray(cmd))

    def write_data(self, data):
        self.i2c.writeto_mem(self.address, 0x40, bytearray(data))

    # ssd1306 ---------
    def init_display(self):
        """init commands and value sequence """
        init_cmd = (  # Sequence of commands and values
            SET_DISP_OFF,
            SET_MEM_ADDR_MODE, 0x00,  # horizontal
            SET_DISP_START_LINE,
            SET_SEG_REMAP_A1,
            SET_COM_OUT_DIR,
            SET_DISP_OFFSET, 0x00,
            SET_DISP_CLK_DIV, 0x80,
            SET_PRECHARGE, 0xF1,
            SET_VCOM_DESEL, 0x30,  # 0.83*Vcc
            SET_CHARGE_PUMP, 0x14,
            SET_CONTRAST, 0xFF,  # Maximum
            SET_ENTIRE_ON,
            SET_DISP_NORMAL,
            SET_MUX_RATIO, self.height - 1,
            SET_COM_PIN_CFG, 0x02 if self.height == 32 else 0x12,
        )
        self.write_cmd(init_cmd)

    def show(self):
        """  Show buffer content on screen """
        xpos0 = 0
        xpos1 = self.width - 1
        if self.width == 64:    # displays with width of 64 pixels are shifted by 32
            xpos0 += 32
            xpos1 += 32
        cmd_sequence = (            # show command and value sequence
            SET_COL_ADDR,
            xpos0,
            xpos1,
            SET_PAGE_ADDR,
            xpos0,
            self.pages - 1
            )
        self.write_cmd(cmd_sequence)
        self.write_data(self.buffer)

    def frambuff(self, buffer, width, height, format=fbuff.MVLSB):
        return(fbuff.FrameBuffer(buffer, width, height, format))

    def disp_on(self, val):
        self.write_cmd([SET_DISP_ON if val else SET_DISP_OFF])

    def vscroll(self, dy):
        #xx = (SET_DISP_START_LINE | dy & 0x3f)
        self.write_cmd((SET_DISP_START_LINE | dy & 0x3f))
        #self.write_cmd([xx])

    def contrast(self, val):
        self.write_cmd((SET_CONTRAST, val & 0xff))

    # framebuffer ---------
    def clear(self, color=0):
        self.fb.fill(color)

    def pixel(self, x, y, color):
        self.fb.pixel(x, y, color)

    def text(self, text, x=0, y=0):
        self.fb.fill_rect(x, y, 8 * len(text), 8, 0)
        self.fb.text(text, x, y)

    def smalltext(self, text, x=0, y=0):
        self.fb.fill_rect(x, y, 4 * len(text), 6, 0)
        font454.text(self.fb, text, x, y)

    def line(self, x1, y1, x2, y2):
        self.fb.line(x1, y1, x2, y2)

    def pixelimage(self, pic, xoffset, yoffset):
        self.clear(0)
        for y, row in enumerate(pic):
            row_str = row.split()[0]
            for x, col in enumerate(row_str):
                if col == "1":
                    self.pixel(x + xoffset, y + yoffset, 1)
