# -------------------------------------------------------------------------------
#   ----  www. r i d o l f a . e u ---
#
#
#### *Copyright 2018  Ridolfa.eu / Rolf S.*
#### *
#### * Licensed under the EUPL, Version 1.1 or – as soon they    *
####   will be approved by the European Commission - subsequent
####   versions of the EUPL (the "Licence");
#### * You may not use this work except in compliance with theLicence.
#### * You may obtain a copy of the Licence at:
####   https://joinup.ec.europa.eu/software/page/eupl
#
#   Rolf Schwerhoff:        RSf_220609
#   Project:                50 Solar
# ------------------------------------------------------------------------------
# ESP32 TTGO tools

import config as cfg
from machine import Pin, SoftSPI, SoftI2C
from time import sleep
from lora_upy import LoRa
from oled_ssd1306 import OledI2c


class BoilerTools:
    """
    This class contains only the necessary functions to use the TTGO
     Version for test purposes.
    """

    def __init__(self):
        pass

    def init_spi_i2c(self):
        spi = SoftSPI(baudrate=cfg.LoRaSpi["Baudrate"],
                      polarity=cfg.LoRaSpi["Polarity"],
                      phase=cfg.LoRaSpi["Phase"],
                      sck=Pin(cfg.LoRaSpi["PinSCK"], Pin.OUT),
                      mosi=Pin(cfg.LoRaSpi["PinMOSI"], Pin.OUT),
                      miso=Pin(cfg.LoRaSpi["PinMISO"], Pin.IN))
        spi.init()

        i2c = SoftI2C(scl=Pin(cfg.I2c["PinSCL"]), sda=Pin(cfg.I2c["PinSDA"]), freq=cfg.I2c["Frequency"])
        return spi, i2c,


class OledTools:
    def __init__(self, i2c):
        self.i2c = i2c
        self.disp = OledI2c(self.i2c, cfg.OLed["I2cAddr"], cfg.OLed["PixWidth"], cfg.OLed["PixHeight"])

    def oled_init(self):
        self.disp.init_display()
        self.disp.disp_on(True)
        self.disp.clear(0)

    def oled_pixelimage(self, pic, xoffset, yoffset):
        self.disp.pixelimage(pic, xoffset, yoffset)

    def oled_startup_screen(self):
        self.disp.clear(0)
        self.disp.text(cfg.Oled_StartScreen["TextA"], 2, 10)
        self.disp.show()
        sleep(3)
        self.disp.clear(0)

    def oled_wrap(self, text):
        text_len = len(text)
        wrappedtext = []
        disp_pix_row_count = 0
        for x in range(0, cfg.OLed["CharMax"],  cfg.OLed["RowCharMax"]):
            xend = x + cfg.OLed["RowCharMax"]
            if xend > cfg.OLed["CharMax"]:
                xend = cfg.OLed["CharMax"]
            line = (disp_pix_row_count, text[x:x + cfg.OLed["RowCharMax"]])
            wrappedtext.append(line)
            disp_pix_row_count += cfg.OLed["PixCharHeight"]
            if xend >= text_len:
                break
        return wrappedtext

    def oled_text_wrap(self, text,
                       start_line=0, height_per_line=8, width_per_char=8, start_pixel_each_line=0,
                       clear_first=True, show_now=True, hold_seconds=0):
        self.disp.clear()
        for lines in self.oled_wrap(text):
            self.disp.text(lines[1], 0, lines[0])

    def oled_text(self, text, x=0, y=0, clear=True):
        if clear:
            self.disp.clear(0)
        self.disp.text(text, x, y)
        self.disp.show()

    def oled_scroll(self, stack, text):
        self.disp.clear(0)
        for i in range(2):
            stack[i] = stack[i + 1]
        stack.pop()
        stack.append(text)
        for i, s in enumerate(stack):
            self.disp.text(s, 0, i*10)
        self.disp.show()


class LoRaTools:

    def __init__(self, spi):
        self.pin_nss = Pin(cfg.LoraPins["NSS"], Pin.OUT)
        self.pin_rst = Pin(cfg.LoraPins["Reset"], Pin.OUT, Pin.PULL_UP)
        self.pin_rxirq = Pin(cfg.LoraPins["DIO0"], Pin.IN)
        self.spi = spi
        self.lora_reset()
        self.lora = LoRa(self.spi, self.get_lorapins())
        self.irq_dio0 = None
        self.show_modem_config()

    def get_lorapins(self):
        return self.pin_rst, self.pin_nss, self.pin_rxirq

    def lora_reset(self):
        """ Reset LoRa device, do a reset pulse """
        self.pin_rst.value(1)
        sleep(.01)
        self.pin_rst.value(0)
        sleep(.01)
        self.pin_rst.value(1)
        sleep(.01)

    def cancel_recv(self):
        """ Cancel  pin  interrupt """
        self.pin_rxirq.irq(handler=None)

    def init_recv_mode(self, callback):
        """
        Start the Raspi /ESP/Pico  Rx  interrupt and wait for incoming messages .
        callback is a handle, pointing to the function that will be processed when an interrupt occurs.
        """
        # print("201", "Wait for messages")
        self.irq_dio0 = self.pin_rxirq.irq(handler=callback, trigger=Pin.IRQ_RISING)
        self.lora.set_rx_mode()

    def send(self, tx_data):
        self.lora.send(tx_data)

    def set_recv_mode(self):
        self.lora.set_rx_mode()

    def get_recv_data(self):
        return self.lora.get_rx_data()

    def get_irq(self):
        return self.irq_dio0

    def get_packet_rssi_strength(self):
        return self.lora.get_packet_strength()

################
    def show_register(self, line_no=00, register=0x01, text="--"):
        """ line_no:  line number of the python script"""
        xx = self.lora.read_register(register)
        print(line_no, text, "0x{:02x}\t{}\t0x{:02x}\t{:08b}".format(register, xx, xx, xx))  # decimal, hex, binary

    def read_modem_status(self):
        ms = self.lora.read_modem_status()
        return dict(
            rx_coding_rate=ms >> 5 & 0x03,
            modem_clear=ms >> 4 & 0x01,
            header_info_valid=ms >> 3 & 0x01,
            rx_ongoing=ms >> 2 & 0x01,
            signal_sync=ms >> 1 & 0x01,
            signal_detected=ms >> 0 & 0x01
        )

    def read_irq_flags(self):
        iqf = self.lora.read_irq_flags()
        return dict(
            rx_timeout=iqf >> 7 & 0x01,
            rx_done=iqf >> 6 & 0x01,
            crc_error=iqf >> 5 & 0x01,
            iqvalid_header=iqf >> 4 & 0x01,
            tx_done=iqf >> 3 & 0x01,
            cad_done=iqf >> 2 & 0x01,
            fhss_change_ch=iqf >> 1 & 0x01,
            cad_detected=iqf >> 0 & 0x01,
        )

    def read_operation_mode(self):
        om = self.lora.read_operation_mode()
        return dict(
            LongRangeMode=om >> 7 & 0x01,  # 7
            AccessSharedReg=om >> 6 & 0x01,  # 6
            reserved=om >> 4 & 0x03,  # 5-4
            LowFrequencyModeOn=om >> 3 & 0x01,  # 3
            mode=om >> 0 & 0x07,  # 2-0
        )

    def debug(self, line, irq_dio0=None):
        print(line, self.read_operation_mode())
        print(line, self.read_irq_flags())
        print(line, self.read_modem_status())

    def show_modem_config(self):
        self.show_register(line_no=00, register=0x1d, text="REG_MODEM_CONFIG_1")
        self.show_register(line_no=00, register=0x1e, text="REG_MODEM_CONFIG_2")
        self.show_register(line_no=00, register=0x26, text="REG_MODEM_CONFIG_3")
